#include "toglmatrix.h"

GLvoid toglVertexCopy3fv(float output[3], float vector[3])
{
	output[0] = vector[0];
	output[1] = vector[1];
	output[2] = vector[2];
}

GLvoid toglVertexCopy4fv(float output[4], float vector[4])
{
	output[0] = vector[0];
	output[1] = vector[1];
	output[2] = vector[2];
	output[3] = vector[3];
}

GLvoid toglVertexCopy3dv(double output[3], double vector[3])
{
	output[0] = vector[0];
	output[1] = vector[1];
	output[2] = vector[2];
}

GLvoid toglVertexCopy4dv(double output[4], double vector[4])
{
	output[0] = vector[0];
	output[1] = vector[1];
	output[2] = vector[2];
	output[3] = vector[3];
}

GLboolean toglVertexNormalize3fv(float output[3], float vector[3])
{
	double	len;


	len = sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);

	if (len == 0.0) {
		toglVertexCopy3fv(output, vector);

		return GL_FALSE;
	}

	output[0] = vector[0] / len;
	output[1] = vector[1] / len;
	output[2] = vector[2] / len;

	return GL_TRUE;
}

GLboolean toglVertexNormalize4fv(float output[4], float vector[4])
{
	double	len;


	if (vector[3] == 0.0f) {
		toglVertexCopy4fv(output, vector);

		return GL_FALSE;
	}

	len = sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);

	if (len == 0.0) {
		toglVertexCopy4fv(output, vector);

		return GL_FALSE;
	}

	output[0] = vector[0] / vector[3] / len;
	output[1] = vector[1] / vector[3] / len;
	output[2] = vector[2] / vector[3] / len;
	output[3] = 1.0f;

	return GL_TRUE;
}

GLboolean toglVertexNormalize3dv(double output[3], double vector[3])
{
	double	len;


	len = sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);

	if (len == 0.0) {
		output[0] = vector[0];
		output[1] = vector[1];
		output[2] = vector[2];

		return GL_FALSE;
	}

	output[0] = vector[0] / len;
	output[1] = vector[1] / len;
	output[2] = vector[2] / len;

	return GL_TRUE;
}

GLboolean toglVertexNormalize4dv(double output[4], double vector[4])
{
	double	len;


	if (vector[3] == 0.0f) {
		output[0] = vector[0];
		output[1] = vector[1];
		output[2] = vector[2];
		output[3] = vector[3];

		return GL_FALSE;
	}

	len = sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);

	if (len == 0.0) {
		output[0] = vector[0];
		output[1] = vector[1];
		output[2] = vector[2];
		output[3] = vector[3];

		return GL_FALSE;
	}

	output[0] = vector[0] / vector[3] / len;
	output[1] = vector[1] / vector[3] / len;
	output[2] = vector[2] / vector[3] / len;
	output[3] = 1.0;

	return GL_TRUE;
}

GLboolean toglMatrixInversefv(float output[16], float matrix[16])
{
	double	epsinv = 0.000001f;
	int		i, j, k;
	double	b, t;
	int		m[4], l[4];
	double	r[4][4];


	for (i = 0; i < 4; i++) {
		for (j = 0; j < 4; j++) {
			r[i][j] = matrix[i * 4 + j];
		}
	}

	for (k = 0; k < 4; k++) {
		m[k] = l[k] = k; 
		b = r[k][k];
		for (j = k; j < 4; j++) {
			for (i = k; i < 4; i++) {
				if (fabs(r[j][i]) > fabs(b)) {
					b = r[j][i];
					l[k] = i;
					m[k] = j;
				}
			}
		}
		if (fabs(b) < epsinv) {
			return GL_FALSE;
		}
		j = l[k];

		if (j > k) {
			for (i = 0; i < 4; i++) {
				t = -r[i][k];
				r[i][k] = r[i][j];
				r[i][j] = t;
			}
		}
		i = m[k];
		if (i > k) {
			for (j = 0; j < 4; j++) {
				t = -r[k][j];
				r[k][j] = r[i][j];
				r[i][j] = t;
			}
		}
		for (i = 0; i < 4; i++) {
			if (i != k) {
				r[k][i] /= -b;
			}
		}
		for (i = 0; i < 4; i++) {
			for (j = 0; j < 4; j++) {
				if (i != k && j != k) {
					r[j][i] += r[k][i] * r [j][k];
				}
			}
		}
		for (j = 0; j < 4; j++) {
			if (j != k) {
				r[j][k] /= b;
			}
		}
		r[k][k] = 1.0 / b;
	}
	for (k = 4 - 2; k >= 0; k--) {
		i = l[k];
		if (i > k) {
			for (j = 0; j < 4; j++) {
				t = r[k][j];
				r[k][j] = -r[i][j];
				r[i][j] = t;
			}
		}
		j = m[k];
		if (j > k) {
			for (i = 0; i < 4; i++) {
				t = r[i][k];
				r[i][k] = -r[i][j];
				r[i][j] = t;
			}
		}
	}
	
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 4; j++) {
			output[i * 4 + j] = r[i][j];
		}
	}

	return GL_TRUE;
}

GLvoid toglMatrixMultMV3fv(float output[3], float matrix[16], float vector[3])
{
	output[0] = 
		matrix[0] * vector[0] +
		matrix[4] * vector[1] +
		matrix[8] * vector[2] +
		matrix[12];
	output[1] = 
		matrix[1] * vector[0] +
		matrix[5] * vector[1] +
		matrix[9] * vector[2] +
		matrix[13];
	output[2] = 
		matrix[2] * vector[0] +
		matrix[6] * vector[1] +
		matrix[10] * vector[2] +
		matrix[14];
}

GLvoid toglMatrixMultMV4fv(float output[4], float matrix[16], float vector[4])
{
	output[0] = 
		matrix[0] * vector[0] +
		matrix[4] * vector[1] +
		matrix[8] * vector[2] +
		matrix[12] * vector[3];
	output[1] = 
		matrix[1] * vector[0] +
		matrix[5] * vector[1] +
		matrix[9] * vector[2] +
		matrix[13] * vector[3];
	output[2] = 
		matrix[2] * vector[0] +
		matrix[6] * vector[1] +
		matrix[10] * vector[2] +
		matrix[14] * vector[3];
	output[3] = 
		matrix[3] * vector[0] +
		matrix[7] * vector[1] +
		matrix[11] * vector[2] +
		matrix[15] * vector[3];
}

GLvoid toglMatrixMultMMfv(float output[16], float mat1[16], float mat2[16])
{
	//  |  0   4   8  12 |
	//  |  1   5   9  13 |
	//  |  2   6  10  14 |
	//  |  3   7  11  15 |

	output[0] = mat1[0] * mat2[0] + mat1[4] * mat2[1] + mat1[8] * mat2[2] + mat1[12] * mat2[3];
	output[1] = mat1[1] * mat2[0] + mat1[5] * mat2[1] + mat1[9] * mat2[2] + mat1[13] * mat2[3];
	output[2] = mat1[2] * mat2[0] + mat1[6] * mat2[1] + mat1[10] * mat2[2] + mat1[14] * mat2[3];
	output[3] = mat1[3] * mat2[0] + mat1[7] * mat2[1] + mat1[11] * mat2[2] + mat1[15] * mat2[3];

	output[4] = mat1[0] * mat2[4] + mat1[4] * mat2[5] + mat1[8] * mat2[6] + mat1[12] * mat2[7];
	output[5] = mat1[1] * mat2[4] + mat1[5] * mat2[5] + mat1[9] * mat2[6] + mat1[13] * mat2[7];
	output[6] = mat1[2] * mat2[4] + mat1[6] * mat2[5] + mat1[10] * mat2[6] + mat1[14] * mat2[7];
	output[7] = mat1[3] * mat2[4] + mat1[7] * mat2[5] + mat1[11] * mat2[6] + mat1[15] * mat2[7];

	output[8] = mat1[0] * mat2[8] + mat1[4] * mat2[9] + mat1[8] * mat2[10] + mat1[12] * mat2[11];
	output[9] = mat1[1] * mat2[8] + mat1[5] * mat2[9] + mat1[9] * mat2[10] + mat1[13] * mat2[11];
	output[10] = mat1[2] * mat2[8] + mat1[6] * mat2[9] + mat1[10] * mat2[10] + mat1[14] * mat2[11];
	output[11] = mat1[3] * mat2[8] + mat1[7] * mat2[9] + mat1[11] * mat2[10] + mat1[15] * mat2[11];

	output[12] = mat1[0] * mat2[12] + mat1[4] * mat2[13] + mat1[8] * mat2[14] + mat1[12] * mat2[15];
	output[13] = mat1[1] * mat2[12] + mat1[5] * mat2[13] + mat1[9] * mat2[14] + mat1[13] * mat2[15];
	output[14] = mat1[2] * mat2[12] + mat1[6] * mat2[13] + mat1[10] * mat2[14] + mat1[14] * mat2[15];
	output[15] = mat1[3] * mat2[12] + mat1[7] * mat2[13] + mat1[11] * mat2[14] + mat1[15] * mat2[15];
}
